/* MayaDate 0.7 for PalmOS 2.0 and greater. */

#include <PalmOS.h>
#include "callback.h"
#include "mayadateRsc.h"


#define mayadatePrefsVersionNum 0.65
#define mayadatePrefID 0x00
#define standardJDCorr 584285
#define mayadateCreator 'Maya'

typedef struct {
    Boolean oldcorr;
} mayaPreferenceType;


typedef struct longdate
{
    UInt16 bak,kat,tun,uin,kin;
} longdate;


void InitDateStuff(void);

UInt32 SavedJulian = 0;
Boolean FirstFlag = true;
UInt32 JD_CORRELATION=584285;


Int16 MayaLoadPrefs(mayaPreferenceType *prefsP) {
    UInt16  prefsSize;
    Int16   prefsVersion = noPreferenceFound;
    
    ErrNonFatalDisplayIf(!prefsP, "null prefP arg");
    
    // Read the preferences / saved-state information.  Fix-up if no
    // prefs or older/newer version
    prefsSize = sizeof (mayaPreferenceType);
    prefsVersion = PrefGetAppPreferences (mayadateCreator, 
					  mayadatePrefID, 
					  prefsP, &prefsSize, true);
    
    // If the preferences version is from a future release (as can
    // happen when going back and syncing to an older version of the
    // device), treat it the same as "not found" because it could be
    // significantly different
    if ( prefsVersion > mayadatePrefsVersionNum )
	prefsVersion = noPreferenceFound;
                
    if ( prefsVersion == noPreferenceFound ) {
	prefsP->oldcorr = false; // default preference is to use new corresp.
    }
    
    return prefsVersion;
}

void MayaSavePrefs (void) {
        mayaPreferenceType prefs;
        
	prefs.oldcorr = true;
	if (JD_CORRELATION == standardJDCorr)
	    prefs.oldcorr = false;
	PrefSetAppPreferences(mayadateCreator, mayadatePrefID, 
			      mayadatePrefsVersionNum, &prefs, 
			      sizeof (mayaPreferenceType), true);
}





// stuff when starting the app
static UInt16 StartApplication(void) {
    FrmGotoForm(MainForm);
    return 0;
}

// stuff when ending the app
static void StopApplication(void) {
    CALLBACK_PROLOGUE
	MayaSavePrefs();
    CALLBACK_EPILOGUE
}

//=======================================================================
// GetObjectPtr
//
// return a pointer to an object with objectID in the form on formP. use
// the active form if formP is NULL.
//=======================================================================

static void* GetObjectPtr( FormPtr formP, UInt16 objectID)
{
        FormPtr frmP;

        if ( formP != NULL )
                frmP = formP;
        else
                frmP = FrmGetActiveForm();
                
        return FrmGetObjectPtr( frmP, FrmGetObjectIndex( frmP, objectID ) );

}       // GetObjectPtr

//=======================================================================
// SetFieldTextFromStr
//
// given a formPtr and fieldID, set the text of the field from the string
// on strP (a copy is made). the active form is used if formP is NULL. works
// even if the field is not editable.
//
// return a pointer to the field.
// 
//=======================================================================
FieldPtr SetFieldTextFromStr( FormPtr formP, UInt16 fieldID, MemPtr strP ) {
    MemHandle       oldTxtH;
    MemHandle       txtH;
    MemPtr          txtP;
    
    FieldPtr        fldP;
    Boolean         fieldWasEditable;
        
    
    // get some space in which to stash the string.
    txtH    = MemHandleNew(  StrLen( strP ) + 1 );
    txtP    = MemHandleLock( txtH );
    
    
    // copy the string.
    StrCopy( txtP, strP );

    
    // unlock the string handle.
    MemHandleUnlock( txtH );
    txtP = 0;
    

    // get the field and the field's current text handle.
    fldP            = GetObjectPtr( formP, fieldID );       
    oldTxtH = FldGetTextHandle( fldP );
    
        
    // set the field's text to the new text.
    fieldWasEditable                = fldP->attr.editable;
    fldP->attr.editable     = true;
    
    FldSetTextHandle( fldP, txtH );
    FldDrawField( fldP );
    
    fldP->attr.editable     = fieldWasEditable;
    
    
    // free the handle AFTER we call FldSetTextHandle().
    if ( oldTxtH != NULL ) MemHandleFree( oldTxtH );
    
    return fldP;
    
}       // SetFieldTextFromStr


// The next routines are from the mtools package by Sean Dwyer.  I
// chopped 'em up a lot but he retains appropriate copyrights under
// the GPL.  

// Get Julian from gregorian date.
UInt32 jul_from_date(UInt16 d, UInt16 m, UInt16 y) {
    Int16 mm, dd, yy, jm;
    double jul, ja, jul2, temp1, temp2;
    
    mm = m; 
    if (mm > 2) {
	yy = y;
	jm = mm + 1;
    } else {
	yy = y - 1;
	jm = mm + 13;
    }
    dd = d;
    temp1 = 365.25 * yy;
    temp2 = (30.6001 * jm) + (dd + 1720995.0);
    jul =  ((UInt32) (temp1)) + ((UInt32)(temp2));
    
    ja = (UInt32) (0.01 * yy);
    jul2 = (UInt32)(jul + (2 - ja + ((UInt32)(0.25 * ja))));
    
    return (UInt32) jul2;
}

void date_from_jul(UInt32 j, UInt16* d, UInt16* m, UInt16* y) {
    double jn;
    UInt32 jalpha,ja,jb,jc,jd,je;
    UInt16 mm, dd, yy;

    jn = j;
    

    // Here's hoping this really can't happen...
    //    if (!check_jul(j)) 
    //  { 
    //    fprintf(stderr, "Bounds error for epoch, invalid data.\n");
    //    return;
    //  }


    jalpha = (UInt32)(((jn - 1867216) - 0.25) / 36524.25);
    ja = (UInt32)(jn + 1 + jalpha - ((int)((0.25 * jalpha))));
    jb = (UInt32)(ja + 1524);
    jc = (UInt32)(6680.0 + ((jb - 2439870.0) - 122.1) / 365.25);
    jd = (UInt32)(365.0 * jc + (0.25 * jc));
    je = (UInt32)((jb - jd) / 30.6001);

    dd = (jb - jd - ((UInt32)( 30.6001 * je ) ));
        
    while (d < 0)
	dd = dd + 7;
        
    mm = je - 1;
    
    if ( mm > 12 )
	mm = mm - 12;

    while (mm < 0)
	mm = mm + 12;
                
    yy = jc - 4715;
    
    if (mm > 2 )
	yy = yy - 1;
    
    *y=yy;
    *m=mm;
    *d=dd;
}


/* try to pick the right mod to use regardless of sign */

double my_mod(double x, double y)
{
        double result;
        
        if ((x < 0) && (y < 0))
                /* both -ve */
                result = (Int32) x % (Int32) y;
        else if ((x > 0) && (y < 0))
                /* x is +ve, y is -ve */
                result = (Int32) x % (Int32) y;
        else if ((x < 0) && (y > 0))
                /* x is -ve, y is +ve */
                result = (Int32) x % (Int32) y ;
        else
                /* both +ve */
                result = (Int32) x % (Int32) y;
        
        return result;
}

/*
Copyright & Disclaimer


(C) 1997,1998, Copyright by Ivan Van Laningham; All Rights Reserved.
Mail to: ivanlan@callware.com

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee or royalty is hereby granted,
provided that the above copyright notice appear in all copies and that both
the copyright notice and this permission notice appear in supporting
documentation or portions thereof, including modifications, that you make.

THE AUTHOR IVAN VAN LANINGHAM DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE!
*/


/* Take a julian date, and convert it to a Long Count */

void jdntolong (UInt32 jdn, longdate *thedate)
{
    double lc;
    double r;
    
    if (jdn > JD_CORRELATION) {
	lc = jdn - JD_CORRELATION;
    } else if (jdn == JD_CORRELATION) {
	lc = 0.0;
    } else {
	lc = jdn - JD_CORRELATION;
    }

    // Under the conditions I'm working with, this Can't Happen(tm),
    // unless it dows.
    //        if (lc < 0.0 ) {
    //  fprintf(stderr, "Bounds error for epoch, invalid data.\n");
    //    return;
    // } else {
    thedate->bak=(UInt16) (lc/144000.0);
    r=my_mod(lc,144000.0);
    thedate->kat=(UInt16) (r/7200);
    r=my_mod(r,7200.0);
    thedate->tun=(UInt16) (r/360);
    r=my_mod(r,360.0);
    thedate->uin=(UInt16) (r/20);
    thedate->kin=(UInt16) my_mod(r,20.0);
}

/* take a julian day number and convert it to a tzolkin_date */

const Char *get_tzolkin_date(UInt32 p, UInt16 *tzolkinnumber, UInt16 *tzolkinday)
{
    double lc=p-JD_CORRELATION;
    UInt16 number, name, ilc;
    static const char *mnth[] =
        { "START","Imix","Ik","Akbal","Kan","Chicchan","Cimi","Manik",
          "Lamat","Muluc","Oc","Chuen","Eb","Ben","Ix","Men","Cib",
          "Caban","Eznab","Cauac","Ahau" };
        
    ilc = (UInt16) lc;
    number = (ilc+9) % 13 +1 ;
    name = (ilc+27) % 20 + 1; 
    
    *tzolkinnumber = number;
    *tzolkinday = name;
    return mnth[name];
}



const Char *get_haab_date(UInt32 x, UInt16 *haabday, UInt16 *haabmo)
{
        UInt32 lc=x-JD_CORRELATION;
        double doh;
        Int16 day, month;
        static const char *mnth[] = 
            { "START","Pop","Uo","Zip","Zotz","Tzec","Xul","Yaxkin","Mol",
              "Chen","Yax","Zac","Ceh","Mac","Kankin","Muan","Pax","Kayab",
              "Cumku","Uayeb" };

        doh = my_mod((lc + 8.0 + 20.0 * (18.0 -1.0)),365.0);
        day = (UInt16) (my_mod(doh,20.0));
        month = (UInt16) (doh/20.0) + 1;

        *haabday = day;
	*haabmo = month;
        return mnth[month];             
}

// Here endeth the inclusion.

// Format the longcount string up all nice
void MyFormatLC(longdate thedate, Char *longdate) {
    Char* baktun[3], katun[3], tun[3], uinal[3], kin[3];

    StrIToA(baktun, thedate.bak);
    StrIToA(katun, thedate.kat);
    StrIToA(tun, thedate.tun);
    StrIToA(uinal, thedate.uin);
    StrIToA(kin, thedate.kin);
    
    StrCopy(longdate, baktun);
    StrCat(longdate, ".");
    StrCat(longdate, katun);
    StrCat(longdate, ".");
    StrCat(longdate, tun);
    StrCat(longdate, ".");
    StrCat(longdate, uinal);
    StrCat(longdate, ".");
    StrCat(longdate, kin);
}
    
void MyDrawGlyph(UInt16 bitmap, UInt16 x, UInt16 y) {
    MemHandle ObjectBitmapHandle;
    BitmapPtr ObjectBitmapPtr;
    
    ObjectBitmapHandle = DmGetResource( bitmapRsc, bitmap);
    ObjectBitmapPtr = MemHandleLock(ObjectBitmapHandle);

    WinDrawBitmap(ObjectBitmapPtr, x, y);

    MemPtrUnlock(ObjectBitmapPtr);
    DmReleaseResource(ObjectBitmapHandle);

}

void MyDrawData(UInt16 dd, UInt16 mm, UInt16 yy) {
    Char* datestring[20], julstring[15], outstring[35], haab[15], 
	haabnum[3], tzolkin[20], tzolnum[3], longcount[20];
    const Char* haabmonth[10];
    const Char* tzolkinname[15];
    UInt16 haabday, haabmo, tzolkinnumber, tzolkinday;
    UInt32 jul;
    longdate thedate;

    MyDrawGlyph(BlankBitmap, 0, 94); 

    DateToAscii(mm, dd, yy, dfMDYWithSlashes, datestring);
    
    // Get the Julian
    jul = jul_from_date(dd, mm, yy);
    StrIToA(julstring, jul); 

    // Format the date/julain line FIXME
    //    StrCopy(outstring, datestring);
    //StrCat(outstring, " -- ");
    //StrCat(outstring, julstring);

    // Get the Long Count
    jdntolong(jul, &thedate);
    MyFormatLC(thedate, longcount);
    // Get the Tzolkin
    *tzolkinname = get_tzolkin_date(jul, &tzolkinnumber, &tzolkinday);
    StrIToA(tzolnum, tzolkinnumber);
    StrCopy(tzolkin, tzolnum);
    StrCat(tzolkin, " ");
    StrCat(tzolkin, *tzolkinname);
    MyDrawGlyph((100+tzolkinnumber), 3, 95);
    MyDrawGlyph((200+tzolkinday), 20, 95);

    // Get the Haab
    *haabmonth = get_haab_date(jul, &haabday, &haabmo);
    StrIToA(haabnum, haabday);
    StrCopy(haab, haabnum);
    StrCat(haab, " ");
    StrCat(haab, *haabmonth);
    MyDrawGlyph((100+haabday), 83, 95);
    MyDrawGlyph((300+haabmo), 100, 95);

    //Print things in appropriate places.
    SetFieldTextFromStr(FrmGetActiveForm(), DateField, datestring);
    SetFieldTextFromStr(FrmGetActiveForm(), JulianField, julstring);
    SetFieldTextFromStr(FrmGetActiveForm(), LongCountField, longcount);
    SetFieldTextFromStr(FrmGetActiveForm(), TzolkinField, tzolkin);
    SetFieldTextFromStr(FrmGetActiveForm(), HaabField, haab);

    SavedJulian = jul;
}    

void InitDateStuff(void) {
    UInt32 seconds;
    UInt16 uday, umonth, uyear;
    DateTimeType date;
    mayaPreferenceType prefs;
    Int16 prefsVersion;
    
    seconds=TimGetSeconds();
    TimSecondsToDateTime(seconds, &date);

    umonth = date.month;
    uday = date.day;
    uyear = date.year;

    prefsVersion = MayaLoadPrefs(&prefs);
    if (prefs.oldcorr)
	JD_CORRELATION = JD_CORRELATION - 2;

    SavedJulian = 0 + jul_from_date(date.day, date.month, date.year);
    //MyDrawData(uday, umonth, uyear);
}

void DoFirstDraw(void) {
    UInt16 m, d, y;
    date_from_jul(SavedJulian, &d, &m, &y);
    MyDrawData(d, m, y);
}

void CopyToClipboard(UInt32 jul) {
    longdate thedate;
    Char* datestring[20], julstring[15], outstring[70], haabnum[3], 
	tzolnum[3], longcount[20];
    const Char* haabmonth[10];
    const Char* tzolkinname[15];
    UInt16 haabday, haabmo, tzolkinnumber, tzolkinday;
    


    jdntolong(jul, &thedate);
    MyFormatLC(thedate, longcount);
    StrCopy(outstring, longcount);
    StrCat(outstring, " ");
    // Get the Tzolkin
    *tzolkinname = get_tzolkin_date(jul, &tzolkinnumber, &tzolkinday);
    StrIToA(tzolnum, tzolkinnumber);
    StrCat(outstring, tzolnum);
    StrCat(outstring, " ");
    StrCat(outstring, *tzolkinname);

    // Get the Haab
    *haabmonth = get_haab_date(jul, &haabday, &haabmo);
    StrIToA(haabnum, haabday);
    StrCat(outstring, haabnum);
    StrCat(outstring, " ");
    StrCat(outstring, *haabmonth);

    ClipboardAddItem(clipboardText, &outstring, StrLen(outstring));

}
    

static Boolean DoPrefs(Boolean initialValue) {
    FormPtr previousForm = FrmGetActiveForm();
    FormPtr frm = FrmInitForm(PrefsForm);
    UInt16 hitButton;
    ControlPtr checkbox = FrmGetObjectPtr(frm, 
					  FrmGetObjectIndex(frm, 
							    CorrelationCheck));
    Boolean newValue;
    FrmSetActiveForm(frm);
    CtlSetValue(checkbox, initialValue);
    
    hitButton = FrmDoDialog(frm);
    
    newValue = CtlGetValue(checkbox);

    if (previousForm)
	FrmSetActiveForm(previousForm);
    
    FrmDeleteForm(frm);
    
    if (hitButton == CorrOKButton)
	return newValue;
    else
	return initialValue;
}
    
    


static Boolean MainFormHandleMenuEvent(UInt16 menuID) {
    Boolean handled = false;
    Boolean check;

    switch (menuID) {
    case OptionsAbout:
        FrmAlert(AboutBoxAlert);
        handled = true;
        break;
    case OptionsPrefs:
	check = DoPrefs(JD_CORRELATION != standardJDCorr);
	JD_CORRELATION=standardJDCorr;
	if (check)
	    JD_CORRELATION = JD_CORRELATION - 2;
	DoFirstDraw();
	handled=true;
	break;
    case EditToClip:
	CopyToClipboard(SavedJulian);
	handled = true;
    }
    return handled;
}

static Boolean MainFormHandleButtonEvent(EventPtr event) {
    Boolean handled = false;
    UInt32 seconds;
    DateTimeType date;
    UInt16 dd, mm, yy;
    UInt32 jul;
    
    //    seconds=TimGetSeconds();
    //TimSecondsToDateTime(seconds, &date);
    //mm = date.month;
    //dd = date.day;
    //yy = date.year;
    date_from_jul(SavedJulian, &dd, &mm, &yy);
    
    switch (event->data.ctlSelect.controlID) {
	//    case NowButton:
	//handled = true;
	//break;
    case BackADayButton:
	jul = jul_from_date(dd, mm, yy) - 1 ;
	date_from_jul(jul, &dd, &mm, &yy);
	handled = true;
	break;
    case ForwardADayButton:
	jul = jul_from_date(dd, mm, yy) + 1 ;
	date_from_jul(jul, &dd, &mm, &yy);
	handled = true;
	break;

    case GoButton:
	handled=SelectDay(selectDayByDay, &mm, &dd, &yy, "Pick a date");
	break;
    }
    MyDrawData(dd,mm,yy);

    return handled;
}


static Boolean MyFormHandleEvent(EventPtr event) {
    Boolean handled = false;

    CALLBACK_PROLOGUE

	switch (event->eType) {
	case ctlSelectEvent:
	    handled = MainFormHandleButtonEvent(event);
	    break;
	    
	case frmOpenEvent:
	    FrmDrawForm(FrmGetActiveForm());
	    if (FirstFlag) {
		DoFirstDraw();
		FirstFlag = false;
	    }
	    handled = true;
	    break;
	    
	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = MainFormHandleMenuEvent(event->data.menu.itemID);     
	    break;
	}
    
    CALLBACK_EPILOGUE
	return handled;
}

static Boolean ApplicationHandleEvent(EventPtr event){
    FormPtr frm;
    Int16 formId;
    Boolean handled = false;

    if (event->eType == frmLoadEvent) {
	formId = event->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);
    	
	
	switch (formId) {
	case MainForm:
	    FrmSetEventHandler(frm, MyFormHandleEvent);
	    break;
	}
	handled = true;
    }
    return handled;
}

static void EventLoop(void) {
    EventType event;
    UInt16 error;

	    
    do {
	EvtGetEvent(&event, evtWaitForever);
	if (! SysHandleEvent(&event))
	    if (! MenuHandleEvent(0, &event, &error))
		if (! ApplicationHandleEvent (&event))
		    FrmDispatchEvent(&event);
    } while ( event.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) {
    Err err = 0;
    if (launchCode == sysAppLaunchCmdNormalLaunch) {
	if ((err = StartApplication()) == 0) {
	    InitDateStuff();
	    EventLoop();
	    StopApplication();
	}
    }
    return err;
}
